/*****************************************************************************/
/*
                                ProxyTunnel.c

*************
** CAUTION **
*************

THIS MODULE IS TASK-ORIENTED, NOT REQUEST-ORIENTED.


WASD supports the CONNECT method which effectively allows tunnelling of RAW
octets through the proxy server.  This facility is most commonly used to allow
secure SSL connections to be established with hosts on the 'other side' of the
proxy server.  This basic mechanism is also used by WASD to provide an extended
range of tunnelling services.  The term 'RAW' is used here to indicate a raw
8 bit, bidirectional, asynchronous exchange of octets between two entities, as
a protocol family, not necessarily as an application (but can be so).


1. [ServiceProxyTunnel] CONNECT
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A service with this configuration is used as a target for CONNECT proxying
(usually SSL through a firewall).  The client expects an HTTP success (200)
response once the remote connection is established, and HTTP error response if
there is a problem, and once established just relays RAW octets through the
proxy server (classic CONNECT behaviour).

  # HTTPD$SERVICE
  [[http://*:8080]]
  [ServiceProxy]  enabled
  [ServiceProxyTunnel]  connect

  # HTTPD$MAP
  [[*:8080]]
  if (request-method:connect)
     pass *:443 *:443
  endif
  pass "403"

This configuration enables CONNECT processing and limits any connect to SSL
tunneling (i.e. port 443 on the remote system).


2. [ServiceProxyTunnel] RAW
~~~~~~~~~~~~~~~~~~~~~~~~~~~
This allows any raw octet client (e.g. telnet) to connect to the port and by
mapping be tunnelled to another host and port to connect to it's service (e.g.
a telnet service).  The usual HTTP responses associated with CONNECT processing
are not provided.

  # HTTPD$SERVICE
  [[http://*:10023]]
  [ServiceProxy]  enabled
  [ServiceProxyTunnel]  raw

  # HTTPD$MAP
  [[*:10023]]
  if (request-method:connect)
     pass *:0 raw://another.host:23
  endif
  pass "403"

Telnet is used in the example above but the principle equally applies to any
protocol that uses a raw 8 bit, bidirectional, asynchronous exchange of
octets.  Another example might be an SMTP service (port 25).


3. RAW via a CHAINED PROXY
~~~~~~~~~~~~~~~~~~~~~~~~~~
This is basically the same as the RAW tunnel described in 2 above except that
it shows a tunnel being established through an up-stream, chained proxy.  It
relies of that proxy allowing a CONNECT to the destination host and more
importantly port required for the tunnel.  Not all CONNECT proxy will allow
this - it's quite common to restrict CONNECT to port 443.  This chained proxy
configuration is also supported for FIREWALL tunnelling.

  # HTTPD$SERVICE
  [[http://*:10025]]
  [ServiceProxy]  enabled
  [ServiceProxyTunnel]  raw

  # HTTPD$MAP
  [[*:10025]]
  if (request-method:connect)
     pass *:0 raw://another.host:25 proxy=chain=proxy.host:8080
  endif
  pass "403"

Any error in connecting to the chained proxy, making the request, connecting to
the destination, etc. (i.e. any error at all) is not reported.  The network
connection is just dropped.  Use WATCH to establish the cause if necessary.


4. [ServiceProxyTunnel] FIREWALL
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
With this configuration a service expects that the first line of text from the
client contains a host name (or IP address) and optional port (e.g.
"the.host.name" or "the.host.name:23").  This allows a variable destination to
be mapped.  The usual HTTP responses associated with CONNECT processing are not
provided.

  # HTTPD$SERVICE
  [[http://*:10023]]
  [ServiceProxy]  enabled
  [ServiceProxyTunnel]  FIREWALL

  # HTTPD$MAP
  [[*:10023]]
  if (request-method:connect)
     pass *:* raw://*:23
     pass * raw://*:23
  endif
  pass "403"

The pass rules force the supplied domain name (and optional port) to be mapped
to the telnet port (23).  Of course the mapping rules could allow the supplied
port to be mapped into the destination if desired.


5. ENCRYPTED TUNNEL
~~~~~~~~~~~~~~~~~~~
Up to this point the tunnels have merely been through the proxy server.  It is
possible to establish and maintain ENCRYPTED TUNNELS between WASD servers.  SSL
is used for this purpose.  This is slightly more complex as both ends of the
tunnel need to be configured.

                 +------------+             +------------+
  <-unencrypted->| WASD proxy |<-ENCRYPTED->| WASD proxy |<-unencrypted->
                 +------------+             +------------+

This arrangement may be used for any stream-oriented, network protocol between
two WASD systems.  As it uses standard CONNECT requests (over SSL) it MAY also
be possible to be configured between WASD and non-WASD servers.

The following example is going to maintain an encrypted tunnel between WASD
servers running on systems KLAATU and GORT.  It is designed to allow a user on
KLAATU to connect to a specified port using a telnet client, and have a telnet
session created on GORT, tunnelled between the two systems via an SSL encrypted
connection.

Source of tunnel:

  # KLAATU HTTPD$SERVICE
  [[http://*:10023]]
  [ServiceProxy]  enabled
  [ServiceClientSSL]  ENABLED
  [ServiceProxyTunnel]  RAW

  # KLAATU HTTPD$MAP
  [[*:10023]]
  # if the client is on the local subnet
  if (remote-addr:192.168.0.0/24 && request-method:connect)
     pass *:0 https://gort.domain:10443 timeout=none,none,none
  endif
  pass "403"

Destination of tunnel:

  # GORT HTTPD$SERVICE
  [[https://*:10443]]
  [ServiceProxy]  enabled
  [ServiceProxyTunnel]  CONNECT

  # GORT HTTPD$MAP
  [[*:10443]]
  # limit the connection to a specific host
  if (remote-addr:192.168.0.10 && request-method:connect)
     pass *:0 raw://gort.domain:23 timeout=none,none,none
  endif
  pass "403"

When a client connects to the service provided by port 10023 on system KLAATU
the connection is immediately processed using a pseudo CONNECT request header. 
The service on this port is a proxy allowed to initiate SSL connections (client
SSL).  This service is mapped to system GORT port 10443, an SSL service that
allows the CONNECT method (tunnelling).  KLAATU's proxy initiates an SSL
connection with GORT.  When established and the CONNECT request from KLAATU is
received, it is mapped via a raw tunnel (8 bit, etc.) to it's own system  port
23 (the telnet service).  Telnet is in use at both ends while encrypted by SSL
inbetween!  Note the use of network addresses and general fail rules used to
control access to this service, as well as the disabling of timers that might
otherwise shutdown the tunnel.


6. ENCRYPTED TUNNEL WITH AUTHENTICATION
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This arrangement is essentially a variation on example 4.  It provides a
cryptographic authentication of the originator (source) of the tunnel.

Source of tunnel:

  # KLAATU HTTPD$SERVICE
  [[http://*:10023]]
  [ServiceProxy]  enabled
  [ServiceClientSSL]  enabled
  [ServiceProxyTunnel]  RAW
  [ServiceClientSSLcert]  HT_ROOT:[LOCAL]HTTPD.PEM

  # KLAATU HTTPD$MAP
  [[*:10023]]
  # if the client is on the local subnet
  if (remote-addr:192.168.0.0/24 && request-method:connect)
     pass *:0 https://gort.domain:10443 timeout=none,none,none
  endif
  pass "403"

Destination of tunnel:

  # GORT HTTPD$SERVICE
  [[https://*:10443]]
  [ServiceProxy]  enabled
  [ServiceProxyTunnel]  CONNECT
  [ServiceProxyAuth]  ENABLED

  # GORT HTTPD$MAP
  [[*:10443]]
  # we'll be relying on X509 authentication
  if (request-method:connect)
     pass *:0 raw://gort.domain:23 timeout=none,none,none
  endif
  pass "403"

  # GORT HTTPD$AUTH
  [[*:10443]]
  [X509]
  * r+w,param="[VF:OPTIONAL]",~4EAB3CBC735F8C7977EBB41D45737E37

This works by configuring the destination service to insist on proxy
authorization.  The authorization realm is X509 which causes the destination to
demand a certificate from the source.  The fingerprint of this certificate is
checked against the authorization rule before the connection is a allowed to
procede.


USAGE EXAMPLE 1
---------------
This is a real-world example used to allow a Mail client (in this case Mozilla
mail, aka Thunderbird) to connect securely to an IMAP and an SMTP server (which
don't natively support SSL) via a WASD tunnel.

Thunderbird settings:

IMAP ... Server Settings ... Use Secure Connection (SSL) ... checked
(that's the port 993).

Outgoing Server (SMTP) ... Use Secure Connection ... SSL ... checked
(that's the port 465).

WASD configuration:

  # HTTPD$SERVICE
  #
  # SSL->IMAP service
  [[https://*:993]]
  [ServiceProxy]  enabled
  [ServiceProxyTunnel]  RAW
  #
  # SSL->SMTP service
  [[https://*:465]]
  [ServiceProxy]  enabled
  [ServiceProxyTunnel]  RAW

  # HTTPD$MAP
  #
  # SSL->IMAP service
  [[*:993]]
  if (request-method:connect)
     pass *:0 raw://imap.host.name:143
  endif
  pass "403"
  #
  # SSL->SMTP service
  [[*:465]]
  if (request-method:connect)
     pass *:0 raw://smtp.host.name:25
  endif
  pass "403"


VERSION HISTORY
---------------
26-MAY-2007  MGD  ProxyTunnel..() provide for SSL client connections
03-SEP-2006  MGD  ProxyTunnelChainConnect() and ProxyTunnelChainConnectAst()
                  to implement raw tunnelling through an intermediate proxy
10-AUG-2004  MGD  'tunnelling' concept generalises CONNECT method
*/
/*****************************************************************************/

#ifdef WASD_VMS_V6
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60000000
#undef __CRTL_VER
#define __CRTL_VER 60000000
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>

/* VMS related header files */
#include <descrip.h>
#include <dvidef.h>
#include <iodef.h>
#include <psldef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application-related header files */
#include "wasd.h"

#define WASD_MODULE "PROXYTUNNEL"

/******************/
/* global storage */
/******************/

PROXY_ACCOUNTING_STRUCT  *ProxyAccountingPtr;

/********************/
/* external storage */
/********************/

extern char  ErrorSanityCheck[],
             SoftwareID[],
             Utility[];

extern ACCOUNTING_STRUCT  *AccountingPtr;
extern CONFIG_STRUCT  Config;
extern IPADDRESS TcpIpEmptyAddress;
extern MSG_STRUCT  Msgs;
extern PROXY_ACCOUNTING_STRUCT  *ProxyAccountingPtr;
extern WATCH_STRUCT  Watch;

/*****************************************************************************/
/*
************
*** NOTE ***  This function takes a pointer to a request!!!
************

HTTP CONNECT method (allows SSL connections through proxy system).
Parse the host name and optional port from the request.
*/

ProxyTunnelRequestParse (REQUEST_STRUCT *rqptr)

{
   int  in, out;
   PROXY_TASK  *tkptr;
   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (rqptr, FI_LI, WATCH_MOD_PROXY,
                 "ProxyTunnelRequestParse() !&Z", rqptr->MappedPathPtr);

   tkptr = rqptr->ProxyTaskPtr;

   cptr = rqptr->MappedPathPtr;

   /*******************/
   /* WASD tunnelling */
   /*******************/

   /* a pragmatic extension to the usual syntax allowed in CONNECT requests */
   if (strsame (cptr, "http://", 7))
   {
      /* unencrypted octets (expect HTTP response from remote) */
      tkptr->ProxyTunnel = PROXY_TUNNEL_HTTP;
      cptr += 7;
   }
   else
   if (strsame (cptr, "https://", 8))
   {
      /* encrypted octets (expect HTTP response from remote) */
      tkptr->ProxyTunnel = PROXY_TUNNEL_HTTPS;
      cptr += 8;
   }
   else
   if (strsame (cptr, "raw://", 6))
   {
      /* unencrypted octets (expect NO HTTP response from remote)  */
      tkptr->ProxyTunnel = PROXY_TUNNEL_RAW;
      cptr += 6;
   }
   else
   {
      /* unencrypted octets, without CONNECT response */
      tkptr->ProxyTunnel = PROXY_TUNNEL_CONNECT;
   }

   /**************/
   /* accounting */
   /**************/

   /* [service-in][proxy-out] (array indices from zero of course!) */
   in = rqptr->ServicePtr->ProxyTunnel;
   if (in) in--;  /* just in case */
   out = tkptr->ProxyTunnel - 1;
   InstanceMutexLock (INSTANCE_MUTEX_HTTPD);
   ProxyAccountingPtr->TunnelCount[in][out]++;
   InstanceMutexUnLock (INSTANCE_MUTEX_HTTPD);

   /*************************/
   /* get server host name  */
   /*************************/

   zptr = (sptr = tkptr->RequestHostName) + sizeof(tkptr->RequestHostName)-1;
   while (*cptr && *cptr != ':' && *cptr != '/' && sptr < zptr)
      *sptr++ = *cptr++;

   if (sptr >= zptr || (*cptr && *cptr != ':'))
   {
      InstanceGblSecIncrLong (&AccountingPtr->RequestErrorCount);
      rqptr->rqResponse.HttpStatus = 502;
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_REQUEST_FORMAT), FI_LI);
      return (STS$K_ERROR);
   }
   *sptr = '\0';

   /******************************/
   /* get (optional) server port */
   /******************************/

   if (*cptr == ':')
   {
      cptr++;
      if (isdigit(*cptr))
      {
         zptr = (sptr = tkptr->RequestPortString) +
                sizeof(tkptr->RequestPortString)-1;
         while (*cptr && isdigit(*cptr) && sptr < zptr) *sptr++ = *cptr++;
         if (sptr >= zptr)
         {
            InstanceGblSecIncrLong (&AccountingPtr->RequestErrorCount);
            rqptr->rqResponse.HttpStatus = 502;
            ErrorGeneral (rqptr, MsgFor(rqptr,MSG_REQUEST_FORMAT), FI_LI);
            return (STS$K_ERROR);
         }
         *sptr = '\0';
         tkptr->RequestPort = atol (tkptr->RequestPortString);
      }
      else
      {
         tkptr->RequestPort = 443;
         memcpy (tkptr->RequestPortString, "443", 4);
      }
   }
   else
   {
      tkptr->RequestPort = 443;
      memcpy (tkptr->RequestPortString, "443", 4);
   }

   /****************************/
   /* build host name and port */
   /****************************/

   zptr = (sptr = tkptr->RequestHostPort) + sizeof(tkptr->RequestHostPort);
   for (cptr = tkptr->RequestHostName;
        *cptr && sptr < zptr;
        *sptr++ = *cptr++);
   if (sptr < zptr) *sptr++ = ':';
   for (cptr = tkptr->RequestPortString;
        *cptr && sptr < zptr;
        *sptr++ = *cptr++);
   if (sptr >= zptr)
   {
      rqptr->rqResponse.HttpStatus = 500;
      ErrorGeneralOverflow (rqptr, FI_LI);
      return (STS$K_ERROR);
   }
   *sptr = '\0';
   tkptr->RequestHostPortLength = sptr - tkptr->RequestHostPort;

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchDataFormatted ("!&Z !UL !&Z !&Z\n",
                          tkptr->RequestHostName,
                          tkptr->RequestPort, tkptr->RequestPortString,
                          tkptr->RequestHostPort);

   return (SS$_NORMAL);
} 

/****************************************************************************/
/*
When explicitly tunnelling using HTTP or HTTPS use the request header line
("CONNECT host:port HTTP/1.1") to rebuild a request header containing the
original host specification provided in that request line but this time with
the proxied remote host in the "Host:" field.
*/ 

int ProxyTunnelRebuildRequest (PROXY_TASK *tkptr)

{
#define STRCAT(string) \
   for (cptr = string; *cptr && sptr < zptr; *sptr++ = *cptr++);

   char  *cptr, *sptr, *zptr;
   REQUEST_STRUCT  *rqptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(tkptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (WATCHTK(tkptr), FI_LI, WATCH_MOD_PROXY,
                 "ProxyTunnelRebuildRequest()");

   rqptr = tkptr->RequestPtr;

   tkptr->RebuiltRequestPtr = sptr =
      VmGetHeap (rqptr, rqptr->rqHeader.RequestHeaderLength + 512);
   zptr = sptr + rqptr->rqHeader.RequestHeaderLength + 512;

   for (cptr = rqptr->rqHeader.RequestHeaderPtr;
        *cptr && NOTEOL(*cptr) && sptr < zptr;
        *sptr++ = *cptr++);
   STRCAT ("\r\nUser-Agent: ");
   STRCAT (SoftwareID);
   STRCAT ("\r\nHost: ");
   STRCAT (tkptr->ConnectHostPortPtr);
   STRCAT ("\r\nX-Forwarded-For: ");
   STRCAT (rqptr->rqClient.IpAddressString);
   STRCAT ("\r\n\r\n");

   if (sptr >= zptr)
   {
      rqptr->rqResponse.HttpStatus = 500;
      ErrorGeneralOverflow (rqptr, FI_LI);
      return (STS$K_ERROR);
   }
   *sptr = '\0';
   tkptr->RebuiltRequestLength = sptr - tkptr->RebuiltRequestPtr;

   if (WATCHING(tkptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchDataFormatted ("!&Z", tkptr->RebuiltRequestPtr);

   return (SS$_NORMAL);

#undef STRCAT
}

/****************************************************************************/
/*
Begin processing appropriate to the mode of tunneling being employed.  This
includes vanilla HTTP CONNECT method processing as well as where this proxy
chains to another.  Called from ProxyNetHostConnectAst() or
SesolaNetClientConnect().
*/

ProxyTunnelBegin (PROXY_TASK *tkptr)

{
   REQUEST_STRUCT *rqptr;
   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(tkptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (WATCHTK(tkptr), FI_LI, WATCH_MOD_PROXY, "ProxyTunnelBegin()");

   /* well, we've got past connection establishment, consider that something */
   tkptr->ResponseStatusCode = 200;

   rqptr = tkptr->RequestPtr;

   if (IPADDRESS_IS_SET (&tkptr->ChainIpAddress))
   {
      if (rqptr->ServicePtr->ProxyTunnel == PROXY_TUNNEL_RAW ||
          rqptr->ServicePtr->ProxyTunnel == PROXY_TUNNEL_FIREWALL)
      {
         /* tunnel access has to be through a chained proxy server */
         ProxyTunnelChainConnect (tkptr);
      }
      else
      {
         ProxyNetWriteRaw (tkptr, &ProxyTunnelWriteAst,
                           rqptr->rqNet.ReadBufferPtr,
                           rqptr->rqNet.ReadIOsb.Count);

         ProxyNetReadRaw (tkptr, &ProxyTunnelReadAst,
                          tkptr->ResponseBufferPtr,
                          tkptr->ResponseBufferSize);
      }
   }
   else
   if (tkptr->ProxyTunnel == PROXY_TUNNEL_CONNECT)
   {
      /* CONNECT success response needs to be returned to the client */
      ResponseHeader (rqptr, tkptr->ResponseStatusCode,
                      NULL, -1, NULL, NULL);
      rqptr->rqResponse.HeaderSent = true;

      if (rqptr->rqNet.SesolaPtr)
         SesolaNetWrite (rqptr->rqNet.SesolaPtr,
                         &ProxyTunnelConnectResponseAst,
                         rqptr->rqResponse.HeaderPtr,
                         rqptr->rqResponse.HeaderLength);
      else
         NetWriteRaw (rqptr, &ProxyTunnelConnectResponseAst,
                      rqptr->rqResponse.HeaderPtr,
                      rqptr->rqResponse.HeaderLength);
   }
   else
   if (tkptr->ProxyTunnel == PROXY_TUNNEL_HTTP)
   {
      /* rebuild the CONNECT request */
      ProxyTunnelRebuildRequest (tkptr);

      /* send the request on through the tunnel */
      ProxyNetWriteRaw (tkptr, &ProxyTunnelWriteAst,
                        tkptr->RebuiltRequestPtr, tkptr->RebuiltRequestLength);

      /* queue a read direct from the tunneled-to service */
      ProxyNetReadRaw (tkptr, &ProxyTunnelReadAst,
                       tkptr->ResponseBufferPtr, tkptr->ResponseBufferSize);
   }
   else
   if (tkptr->ProxyTunnel == PROXY_TUNNEL_HTTPS)
   {
      /* rebuild the CONNECT request */
      ProxyTunnelRebuildRequest (tkptr);

      /* send the request on through the tunnel */
      SesolaNetWrite (tkptr->SesolaPtr, &ProxyTunnelWriteAst,
                      tkptr->RebuiltRequestPtr, tkptr->RebuiltRequestLength);

      /* queue a read direct from the tunneled-to service */
      SesolaNetRead (tkptr->SesolaPtr, &ProxyTunnelReadAst,
                     tkptr->ResponseBufferPtr, tkptr->ResponseBufferSize);
   }
   else
   if (tkptr->ProxyTunnel == PROXY_TUNNEL_RAW)
   {
      if (rqptr->ServicePtr->ProxyTunnel == PROXY_TUNNEL_FIREWALL ||
          rqptr->ServicePtr->ProxyTunnel == PROXY_TUNNEL_RAW)
      {
         /* queue a read direct from the client */
         if (rqptr->rqNet.SesolaPtr)
            SesolaNetRead (rqptr->rqNet.SesolaPtr,
                           &ProxyTunnelNetReadAst,
                           rqptr->rqNet.ReadBufferPtr,
                           rqptr->rqNet.ReadBufferSize);
         else
            NetReadRaw (rqptr, &ProxyTunnelNetReadAst,
                        rqptr->rqNet.ReadBufferPtr,
                        rqptr->rqNet.ReadBufferSize);

         /* queue a read direct from the tunneled-to service */
         ProxyNetReadRaw (tkptr, &ProxyTunnelReadAst,
                          tkptr->ResponseBufferPtr, tkptr->ResponseBufferSize);
      }
      else
      {
         /* CONNECT success response needs to be returned to the client */
         ResponseHeader (rqptr, tkptr->ResponseStatusCode,
                         NULL, -1, NULL, NULL);
         rqptr->rqResponse.HeaderSent = true;

         /* send that response back to the client */
         if (rqptr->rqNet.SesolaPtr)
            SesolaNetWrite (rqptr->rqNet.SesolaPtr,
                            &ProxyTunnelConnectResponseAst,
                            rqptr->rqResponse.HeaderPtr,
                            rqptr->rqResponse.HeaderLength);
         else
            NetWriteRaw (rqptr, &ProxyTunnelConnectResponseAst,
                         rqptr->rqResponse.HeaderPtr,
                         rqptr->rqResponse.HeaderLength);
      }
   }
   else
      ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);
}

/****************************************************************************/
/*
Generate a CONNECT request header for the request host:port derived by
ProxyTunnelRequestParse() and then write this to the chained proxy server.
*/ 

int ProxyTunnelChainConnect (PROXY_TASK *tkptr)

{
#define STRCAT(string) \
   for (cptr = string; *cptr && sptr < zptr; *sptr++ = *cptr++);

   char  *cptr, *sptr, *zptr;
   REQUEST_STRUCT  *rqptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(tkptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (WATCHTK(tkptr), FI_LI, WATCH_MOD_PROXY,
                 "ProxyTunnelChainConnect()");

   rqptr = tkptr->RequestPtr;

   tkptr->RebuiltRequestPtr = sptr = VmGetHeap (rqptr, 512);
   zptr = sptr + 512;

   STRCAT ("CONNECT ");
   STRCAT (tkptr->RequestHostPort);
   if (tkptr->RequestHttpVersion == HTTP_VERSION_1_1)
      STRCAT (" HTTP/1.1")
   else
      STRCAT (" HTTP/1.0")
   STRCAT ("\r\nUser-Agent: ");
   STRCAT (SoftwareID);
   STRCAT ("\r\nHost: ");
   STRCAT (tkptr->ConnectHostPortPtr);
   STRCAT ("\r\nX-Forwarded-For: ");
   STRCAT (rqptr->rqClient.IpAddressString);
   STRCAT ("\r\n\r\n");

   if (sptr >= zptr)
   {
      rqptr->rqResponse.HttpStatus = 500;
      ErrorGeneralOverflow (rqptr, FI_LI);
      return (STS$K_ERROR);
   }
   *sptr = '\0';
   tkptr->RebuiltRequestLength = sptr - tkptr->RebuiltRequestPtr;

   if (WATCHING(tkptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchDataFormatted ("!&Z", tkptr->RebuiltRequestPtr);

   /* to the CONNECT request write synchonously */
   ProxyNetWriteRaw (tkptr, NULL,
                     tkptr->RebuiltRequestPtr,
                     tkptr->RebuiltRequestLength);

   /* queue a read for the CONNECT response */
   ProxyNetReadRaw (tkptr, &ProxyTunnelChainConnectAst,
                    tkptr->ResponseBufferPtr, tkptr->ResponseBufferSize);

   return (SS$_NORMAL);

#undef STRCAT
}

/****************************************************************************/
/*
The chained proxy server has responded to the CONNECT request.  Check for
network errors and that the response looks something like HTTP success. If it
does begin reading from each of the client and the chained proxy server.
*/

ProxyTunnelChainConnectAst (PROXY_TASK *tkptr)

{
   int  DataCount;
   char  *DataPtr;
   REQUEST_STRUCT  *rqptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(tkptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (WATCHTK(tkptr), FI_LI, WATCH_MOD_PROXY,
                 "ProxyTunnelChainConnectAst() !&F !&S !UL",
                 &ProxyTunnelChainConnectAst,
                 tkptr->ProxyReadIOsb.Status,
                 tkptr->ProxyReadIOsb.Count);

   if (VMSnok (tkptr->ProxyReadIOsb.Status))
   {
      ProxyEnd (tkptr);
      return;
   }

   DataPtr = tkptr->ResponseBufferPtr;
   DataCount = tkptr->ProxyReadIOsb.Count;

   if (WATCHING(tkptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchDataDump (DataPtr, DataCount);

   DataPtr[DataCount] = '\0';
   if (memcmp (DataPtr, "HTTP/1.", 7))
   {
      ProxyEnd (tkptr);
      return;
   }
   DataPtr += 7;
   while (*DataPtr && *DataPtr != ' ') DataPtr++; 
   while (*DataPtr == ' ') DataPtr++; 
   if (memcmp (DataPtr, "200", 3))
   {
      ProxyEnd (tkptr);
      return;
   }

   rqptr = tkptr->RequestPtr;

   /* queue a read direct from the client */
   if (rqptr->rqNet.SesolaPtr)
      SesolaNetRead (rqptr->rqNet.SesolaPtr,
                     &ProxyTunnelNetReadAst,
                     rqptr->rqNet.ReadBufferPtr,
                     rqptr->rqNet.ReadBufferSize);
   else
      NetReadRaw (rqptr, &ProxyTunnelNetReadAst,
                  rqptr->rqNet.ReadBufferPtr,
                  rqptr->rqNet.ReadBufferSize);

   /* queue a read direct from the tunneled-to service */
   ProxyNetReadRaw (tkptr, &ProxyTunnelReadAst,
                    tkptr->ResponseBufferPtr, tkptr->ResponseBufferSize);
}

/****************************************************************************/
/*
************
*** NOTE ***  This function takes a pointer to a request!!!
************

A standard CONNECT response has been sent to the client by this server.
It is necessary to provide this explicitly with a 'raw' connection.
Check status.  If OK then begin to tunnel.
*/

ProxyTunnelConnectResponseAst (REQUEST_STRUCT *rqptr)

{
   PROXY_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (rqptr, FI_LI, WATCH_MOD_PROXY,
                 "ProxyTunnelConnectResponseAst() !&F !&S !UL",
                 &ProxyTunnelConnectResponseAst,
                 rqptr->rqNet.WriteIOsb.Status,
                 rqptr->rqNet.WriteIOsb.Count);

   tkptr = rqptr->ProxyTaskPtr; 

   if (VMSnok (rqptr->rqNet.WriteIOsb.Status))
   {
      ProxyEnd (tkptr);
      return;
   }

   /* queue a read direct from the client */
   if (rqptr->rqNet.SesolaPtr)
      SesolaNetRead (rqptr->rqNet.SesolaPtr,
                     &ProxyTunnelNetReadAst,
                     rqptr->rqNet.ReadBufferPtr,
                     rqptr->rqNet.ReadBufferSize);
   else
      NetReadRaw (rqptr, &ProxyTunnelNetReadAst,
                  rqptr->rqNet.ReadBufferPtr,
                  rqptr->rqNet.ReadBufferSize);

   /* queue a read direct from the tunneled-to service */
   if (tkptr->SesolaPtr)
      SesolaNetRead (tkptr->SesolaPtr, &ProxyTunnelReadAst,
                     tkptr->ResponseBufferPtr, tkptr->ResponseBufferSize);
   else
      ProxyNetReadRaw (tkptr, &ProxyTunnelReadAst,
                      tkptr->ResponseBufferPtr, tkptr->ResponseBufferSize);
}

/****************************************************************************/
/*
************
*** NOTE ***  This function takes a pointer to a request!!!
************

A read from the client has completed.
Check status.  If OK write it to the remote connection.
*/

ProxyTunnelNetReadAst (REQUEST_STRUCT *rqptr)

{
   PROXY_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (rqptr, FI_LI, WATCH_MOD_PROXY,
                 "ProxyTunnelNetReadAst() !&F !&S !UL",
                 &ProxyTunnelNetReadAst,
                 rqptr->rqNet.ReadIOsb.Status,
                 rqptr->rqNet.ReadIOsb.Count);

   tkptr = rqptr->ProxyTaskPtr; 

   if (VMSnok (rqptr->rqNet.ReadIOsb.Status))
   {
      ProxyEnd (tkptr);
      return;
   }

   if (tkptr->SesolaPtr)
      SesolaNetWrite (tkptr->SesolaPtr,
                      &ProxyTunnelWriteAst,
                      rqptr->rqNet.ReadBufferPtr,
                      rqptr->rqNet.ReadIOsb.Count);
   else
      ProxyNetWriteRaw (tkptr, &ProxyTunnelWriteAst,
                        rqptr->rqNet.ReadBufferPtr,
                        rqptr->rqNet.ReadIOsb.Count);
}

/****************************************************************************/
/*
************
*** NOTE ***  This function takes a pointer to a request!!!
************

A write to the client has completed.
Check status.  If OK read more from the remote connection.
*/

ProxyTunnelNetWriteAst (REQUEST_STRUCT *rqptr)

{
   PROXY_TASK  *tkptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (rqptr, FI_LI, WATCH_MOD_PROXY,
                 "ProxyTunnelNetWriteAst() !&F !&S !UL",
                 &ProxyTunnelNetWriteAst,
                 rqptr->rqNet.WriteIOsb.Status,
                 rqptr->rqNet.WriteIOsb.Count);

   tkptr = rqptr->ProxyTaskPtr; 

   if (VMSnok (rqptr->rqNet.WriteIOsb.Status))
   {
      ProxyEnd (tkptr);
      return;
   }

   if (tkptr->SesolaPtr)
      SesolaNetRead (tkptr->SesolaPtr,
                     &ProxyTunnelReadAst,
                     tkptr->ResponseBufferPtr,
                     tkptr->ResponseBufferSize);
   else
      ProxyNetReadRaw (tkptr, &ProxyTunnelReadAst,
                       tkptr->ResponseBufferPtr,
                       tkptr->ResponseBufferSize);
}

/****************************************************************************/
/*
AST completion of read from remote connection.
Check status.  If OK write it to the client connection.
*/

ProxyTunnelReadAst (PROXY_TASK *tkptr)

{
   int  DataCount;
   char  *cptr, *zptr,
         *DataPtr;
   REQUEST_STRUCT *rqptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(tkptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (WATCHTK(tkptr), FI_LI, WATCH_MOD_PROXY,
                 "ProxyTunnelReadAst() !&F !&S !UL",
                 &ProxyTunnelReadAst,
                 tkptr->ProxyReadIOsb.Status,
                 tkptr->ProxyReadIOsb.Count);

   rqptr = tkptr->RequestPtr;

   if (VMSnok (tkptr->ProxyReadIOsb.Status))
   {
      ProxyEnd (tkptr);
      return;
   }

   DataPtr = tkptr->ResponseBufferPtr;
   DataCount = tkptr->ProxyReadIOsb.Count;

   if ((tkptr->ProxyTunnel == PROXY_TUNNEL_CONNECT ||
        tkptr->ProxyTunnel == PROXY_TUNNEL_HTTP ||
        tkptr->ProxyTunnel == PROXY_TUNNEL_HTTPS) &&
       (rqptr->ServicePtr->ProxyTunnel == PROXY_TUNNEL_FIREWALL ||
        rqptr->ServicePtr->ProxyTunnel == PROXY_TUNNEL_RAW))
   {
      /* need to absorb any CONNECT response header */
      if (tkptr->ResponseConsecutiveNewLineCount < 2)
      {
         zptr = (cptr = DataPtr) + DataCount;
         while (cptr < zptr)
         {
            if (*cptr == '\n')
               tkptr->ResponseConsecutiveNewLineCount++;
            else
            if (*cptr != '\r')
               tkptr->ResponseConsecutiveNewLineCount = 0;
            cptr++;
            if (tkptr->ResponseConsecutiveNewLineCount >= 2) break;
         }
         DataCount -= cptr - DataPtr;
         DataPtr = cptr;

         if (!DataCount)
         {
            /* fudge these and directly deliver the AST */
            rqptr->rqNet.WriteIOsb.Status = SS$_NORMAL;
            rqptr->rqNet.WriteIOsb.Count = 0;
            SysDclAst (ProxyTunnelNetWriteAst, rqptr);
            return;
         }
      }
   }

  if (rqptr->rqNet.SesolaPtr)
     SesolaNetWrite (rqptr->rqNet.SesolaPtr,
                     &ProxyTunnelNetWriteAst,
                     DataPtr, DataCount);
  else
     NetWriteRaw (rqptr, &ProxyTunnelNetWriteAst,
                  DataPtr, DataCount);
}

/****************************************************************************/
/*
A write to the remote connection has completed. 
Check status.  If OK read more from the client connection.
*/

ProxyTunnelWriteAst (PROXY_TASK *tkptr)

{
   REQUEST_STRUCT *rqptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(tkptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (WATCHTK(tkptr), FI_LI, WATCH_MOD_PROXY,
                 "ProxyTunnelWriteAst() !&F !&S !UL",
                 &ProxyTunnelWriteAst,
                 tkptr->ProxyWriteIOsb.Status,
                 tkptr->ProxyWriteIOsb.Count);

   rqptr = tkptr->RequestPtr;

   if (VMSnok (tkptr->ProxyWriteIOsb.Status))
   {
      ProxyEnd (tkptr);
      return;
   }

   if (rqptr->rqNet.SesolaPtr)
      SesolaNetRead (rqptr->rqNet.SesolaPtr,
                     &ProxyTunnelNetReadAst,
                     rqptr->rqNet.ReadBufferPtr,
                     rqptr->rqNet.ReadBufferSize);
   else
      NetReadRaw (rqptr, &ProxyTunnelNetReadAst,
                  rqptr->rqNet.ReadBufferPtr,
                  rqptr->rqNet.ReadBufferSize);
}

/****************************************************************************/

